#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import os
import re
import shutil
from os.path import join, relpath

from .template import DumperTemplate


__all__ = [
    "Autoruns",
    "Clipboard",
    "CommandLines",
    "CriticalProcessesInfo",
    "Devices",
    "DumpInfo",
    "FilesList",
    "Kernel",
    "LsaSecrets",
    "Malfind",
    "Mimikatz",
    "NetworkConnections",
    "ProcessesInfo",
    "Registry",
    "Timeline",
    "UserActivities",
    "UserHashes",
]


class Autoruns(DumperTemplate):
    """
    Dumper for mapping AutoStart Extensibility Points to running processes.
    """
    def run(self):
        self.commands('autoruns')


class Clipboard(DumperTemplate):
    """
    Dumper for collecting the content of the clipboard.
    """
    messages = ["Something was found in the clipboard ; you should take a look at it, there could be some passwords..."]
    
    def run(self):
        out = self.call("clipboard", failmode="warn").split('\n')
        # we want to know if column "Data" has a value, so take its start index
        n = out[0].index("Data")
        new = out[:2]
        for i in range(2, len(out)):
            # now, take this line only if column "Data" has a value
            if len(out[i].rstrip()) > n:
                new.append(out[i])
        r = self.result("content", "txt")
        return self.save('\n'.join(new), r, header=2)


class CommandLines(DumperTemplate):
    """
    Dumper for collecting last commands ran.
    """
    def run(self):
        self.commands('cmdscan', 'consoles', 'cmdline')


class CriticalProcessesInfo(DumperTemplate):
    """
    Dumper for checking critical processes.
    """
    # TODO: check that lsass.exe is only one instance AND its parent is wininit.exe (Vista+) or winlogon.exe (XP-)
    def run(self):
        psscan, psxview = self.commands('psscan', 'psxview', save=False, options="--output=greptext")
        info = ""
        # check for single instance of lsass.exe and services.exe
        for ps in ["lsass.exe", "services.exe"]:
            name = ps.split(".", 1)[0]
            instances = []
            for l in psscan.strip().split('\n'):
                l = l.split("|")
                if ps in l:
                    instances += [(l[3], l[4])]  # PID, PPID
            pids = list(map(lambda x: x[0], instances))
            for l in psxview.strip().split('\n'):
                l = l.split("|")
                pid = l[3]
                if ps in l and pid not in pids:
                    instances += [(pid, "N/A")]  # PID, PPID
            if len(instances) > 1:
                info += "Multiple instances of {} found".format(ps)
                for pid, ppid in sorted(instances, key=lambda x: int(x[0])):
                    info += "\n{} (parent: {})".format(pid, ppid)
                self.save(info, self.result("multiple-{}".format(name)))


class Devices(DumperTemplate):
    """
    Dumper for listing device drivers.
    """
    def run(self):
        self.commands('devicetree')


class DumpInfo(DumperTemplate):
    """
    Dumper of general information about the memory image.
    """
    def run(self):
        """
        Executes a series of informational Volatility commands.
        """
        self.commands('crashinfo', 'hibinfo', 'vboxinfo', 'vmwareinfo')


class FilesList(DumperTemplate):
    """
    Dumper for the list of file objects.
    """
    def run(self):
        self.commands('filescan')


class Kernel(DumperTemplate):
    """
    Dumper for getting kernel-related information.
    """
    def run(self):
        self.commands('callbacks', 'timers')


class LsaSecrets(DumperTemplate):
    """
    Dumper for collecting user activity artifacts.
    """
    def run(self):
        """
        Executes 'hivelist' and 'lsadump' Volatility commands in order to dump LSA secrets.
        """
        sec, sys = None, None
        for line in self.call("hivelist").split('\n'):
            if line.startswith("0x"):
                start, end, hive = line.split(" ", 2)
                if re.search(r'(security)$', hive, re.I) is not None:
                    sec = start
                if re.search(r'(system)$', hive, re.I) is not None:
                    sys = start
        if sec is not None and sys is not None:
            self.commands("lsadump", options="-y {} -s {}".format(sys, sec))


class Malfind(DumperTemplate):
    """
    Dumper for finding hidden or injected code in user mode memory, unlinked DLLs and detecting hollowing techniques.
    """
    def run(self):
        if self.commands('malfinddeep', failmode="warn")[0] is None:
            self.commands('malfind')
        self.commands('ldrmodules', 'hollowfind')


class Mimikatz(DumperTemplate):
    """
    Dumper for collecting user passwords present in memory.
    """
    def run(self):
        self.commands('mimikatz', header=2, failmode="warn")


class NetworkConnections(DumperTemplate):
    """
    Dumper for scanning TCP connections and sockets.
    """
    def run(self):
        self.commands('netscan', 'connscan', 'sockscan')


class ProcessesInfo(DumperTemplate):
    """
    Dumper for listing processes.
    """
    def run(self):
        self.commands('pslist', 'pstree', 'psscan', 'psxview')


class Registry(DumperTemplate):
    """
    Dumper for extracting all available registry hives.
    """
    def run(self):
        d = self.dump.out_dir
        tmp = os.path.join(d, ".registry")
        try:
            os.makedirs(tmp)
        except OSError:
            pass
        self.commands('dumpregistry', options="--dump-dir {}".format(tmp))
        for fn in os.listdir(tmp):
            fp = join(tmp, fn)
            with open(fp, 'rb') as f:
                _ = f.read().strip('\x00')
            if len(_) > 0:
                t = fn.split('.')
                nf = "{}.{}".format('-'.join(t[:-1]), t[-1])
                np = join(d, nf)
                with open(np, 'wb') as f:
                    f.write(_)
                self.logger.info("> {}".format(relpath(np)))
        shutil.rmtree(tmp)


class Timeline(DumperTemplate):
    """
    Dumper for creating a timeline from various artifacts in memory.
    """
    def run(self):
        out = '\n'.join(self.commands('timeliner', ('mftparser', "-C --output=body"), 'shellbags', save=False,
                                      options="--output=body", failmode="warn"))
        tmp = self.save(out, self.result("temp"), verbose=False)
        out, _ = self.shell("mactime -d -b {}".format(tmp), True)
        self.save(out, self.result("mactimes", "csv"))


class UserActivities(DumperTemplate):
    """
    Dumper for collecting user activity artifacts.
    """
    def run(self):
        self.commands('userassist')


class UserHashes(DumperTemplate):
    """
    Dumper of Windows user hashes from the memory image.
    """
    def run(self):
        """
        Executes 'hivelist' and 'hashdump' Volatility commands in order to dump user hashes.
        """
        sam, sys = None, None
        for line in self.call("hivelist").split('\n'):
            if line.startswith("0x"):
                start, end, hive = line.split(" ", 2)
                if re.search(r'(sam)$', hive, re.I) is not None:
                    sam = start
                if re.search(r'(system)$', hive, re.I) is not None:
                    sys = start
        if sam is not None and sys is not None:
            self.commands("hashdump", options="{} -s {}".format(sam, sys))




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import logging
import os
import re
import shutil
import string
from collections import deque
from os.path import dirname, exists, isdir, isfile, join, relpath, splitext
from subprocess import CalledProcessError, Popen, PIPE


__all__ = ["DumperTemplate"]
logger = logging.getLogger("main")


class DumperTemplate(object):
    """
    Dumper template class for handling base operations on the memory dump for handled applications. It can execute
     Volatility commands using the attached VolatilityMemDump instance and carve files using Foremost.
    """
    _predef_messages = [
        "Raw data (.data files) requires manual handling ;\n"
        "Follow this procedure:\n"
        " 1. Open the collected resources with Gimp\n"
        " 2. Set the width and height to the expected screen resolution\n"
        " 3. Set another color palette than 'RVB'\n"
        "Restart this procedure by setting other parameters for width|height"
        "|palette.",
        "VAD objects (collected with 'vaddump' require further research ;\n"
        "If you know some keywords that could match text elements, you can grep"
        " the .dmp objects.",
    ]
    messages = []
    procnames = None  # by default, the GenericDumper does not correspond to any particular process name, meaning that a
                      #  dumper inheriting this class without redefining procnames will be processed regardless of the
                      #  process list
    only_parent = False

    def __init__(self, dump, pids):
        """
        Simple constructor for setting base attributes.

        :param dump: VolatilityMemDump instance
        :param pids: list of pids corresponding to the application name in dump
        """
        if not isinstance(pids, list) or any(not x.isdigit() for x in pids):
            raise ValueError("Bad list of PID's (should be list of digits)")
        self.dump = dump
        self.logger = logger
        self.name = self.__class__.__name__
        self.pids = list(set(pids))
        self._filter_pids()
        self.__nopid = []
        if not hasattr(self.__class__, "missing_plugins"):
            self.__class__.missing_plugins = []
    
    def _filter_pids(self):
        """
        Filter out child PID's of processes whose own and parent procnames are in self.procnames.
        """
        if self.only_parent:
            for k, v in self.ppids.items():
                ppid, ppname = k
                for pid, pname in v:
                    if ppname == pname:
                        try:
                            self.pids.remove(pid)
                        except ValueError:
                            continue

    def _is_processed(self, pid=None):
        """
        Check if this dumper was already processed for the given PID relying on a cache file.
         
        :param pid: PID (if any)
        :return:    True if dumper already processed
        """
        if self._update:
            with open(self.dump._cachefile) as f:
                dumpers = f.read().splitlines()[1:]
            name = self.name if pid is None else "{}-{}".format(self.name, pid)
            if name in dumpers:
                self.logger.debug("> Already processed ; nothing to update")
                return True
        return False

    def _makedir(self, *subdirs):
        """
        Compose the dump directory path and create it if it does not exist.

        :param *subdirs: string pieces of the dump directory path
        :return:         the dump directory path as a single string
        """
        if any(not isinstance(x, str) for x in subdirs):
            raise ValueError("Bad list of subdirectories (should be a list of str)")
        dst = join(self.dump.out_dir, *subdirs)
        if not isdir(dst):
            os.makedirs(dst)
        return dst

    def _run(self, update=False):
        """
        Consecutively execute the public 'run()' method on each PID from the pids list.
        
        :param update: update output directory (meaning that collected pieces of information should not be recollected
                        using Volatility commands)
        :return:       list of messages (if any) associated to this dumper
        """
        self._update, n = update, self.name
        ok = False
        # applies to system dumpers
        if len(self.pids) == 0:
            if not self._is_processed():
                logger.debug("Dumping information...")
                ok = self.run() is not None
                self.dump._update_cache(dumper=n)
        else:
            for pid in self.pids:
                if not self._is_processed(pid):
                    logger.debug("Dumping for PID {}...".format(pid))
                    self.dump.config.PID = pid
                    ok = self.run() is not None or ok
                    self.dump._update_cache(dumper="{}-{}".format(n, pid))
        return self.messages if ok else []
    
    @property
    def pid(self):
        """
        Dummy shortcut property to self.dump.config.PID.
        """
        _ = self.dump.config.PID
        return _ if _ in self.pids else None
    
    @property
    def ppids(self):
        """
        Dummy shortcut property to self.dump.ppids.
        """
        return self.dump.ppids
    
    def call(self, *args, **kwargs):
        """
        Dummy shortcut to self.dump.call.
        """
        return self.dump.call(*args, **kwargs)

    def carve(self, *types, **kwargs):
        """
        Carve files of given types from the input resource with Foremost.

        :param types: list of extensions to be carved from the resource
        :param clean: remove dump file after processing
        """
        clean = kwargs.pop('clean', False)
        if any(not isinstance(x, str) for x in types):
            raise ValueError("Bad list of types (should be a list of str)")
        if not isinstance(clean, bool):
            raise ValueError("Bad clean variable type (should be bool)")
        fp = self.memdump(verbose=False)
        if clean:
            self.dump._artifacts.append(fp)
        logger.debug("> Carving with Foremost...")
        folder, _ = splitext(fp)
        cmd = "foremost {} -o {} {}".format(fp, folder, ["", "-t {}".format(",".join(types))][len(types) > 0])
        logger.debug(">> {}".format(cmd))
        self.shell(cmd)
        with open(join(folder, "audit.txt")) as f:
            disp = False
            for line in f.readlines():
                line = line.strip()
                if "Finish:" in line:
                    break
                if all([x in line for x in ("Num", "Name", "Size", "Comment")]):
                    disp = True
                if disp and len(line) > 0 and ":" in line:
                    fn = line.split(":", 1)[1].strip().split()[0]
                    fp = join(folder, splitext(fn)[1][1:], fn)
                    logger.info("> {}".format(relpath(fp)))
    
    def commands(self, *cmds, **kwargs):
        """
        Call multiple Volatility commands on the associated dump with the given options and save the results.
        """
        fmt = kwargs.pop('fmt', None)
        options = kwargs.pop('options', None)
        save = kwargs.pop('save', True)
        verbose = kwargs.pop('verbose', True)
        header = kwargs.pop('header', 0)
        res = []
        for cmd in cmds:
            if cmd in self.__class__.missing_plugins:
                continue
            # commands can be defined as 'command' or '(command, options)'
            if isinstance(cmd, tuple):
                cmd, opt = cmd
            else:
                opt = None or options
            # handle options separately
            out = self.call(cmd, opt, **kwargs)
            if out is None:
                self.__class__.missing_plugins.append(cmd)
                res.append(None)
                continue
            if save:
                # adapt output format if necessary
                s = "--output="
                if fmt is None and opt is not None and s in opt:
                    fmt = filter(lambda x: s in x, opt.split())[0].split('=')[1]
                else:
                    fmt = "txt"
                dst = self.result(cmd, fmt)
                res.append(self.save(out, dst, verbose, header))
            else:
                res.append(out)
        return tuple(res)
    
    def execute(self, *args, **kwargs):
        """
        Dummy shortcut to self.dump.execute.
        """
        return self.dump.execute(*args, **kwargs)
    
    def has(self, pid):
        """
        Dummy function to verify if input PID matches dumper's associated one.
        """
        return self.pid == pid

    def memdump(self, verbose=True):
        """
        Executes the 'memdump' Volatility command for extracting the related process memory.

        :param verbose: display the log message after saving the memory dump
        :return:        path to the process memory dump
        """
        if not isinstance(verbose, bool):
            raise ValueError("Bad verbose variable type (should be bool)")
        cmd = 'memdump'
        dst = self.result(cmd, "data")
        if self._update and exists(dst):
            return dst
        out = self.call(cmd, "-p {} --dump-dir {}".format(self.pid, self.dump.out_dir))
        src = join(self.dump.out_dir, "{}.dmp".format(self.pid))
        # could already exist if previous memdump result was cached
        if not exists(dst):
            shutil.move(src, dst)
        if verbose:
            logger.info("> {}".format(relpath(dst)))
        return dst

    def memsearch(self, split_on_nullbyte=False):
        """
        Executes the 'memdump' Volatility command and parses the collected dump against user-defined regex.
        Valid patterns structure:
          self.re_patterns := list(tuples(re_pattern, format, short_descr))
        
        :param split_on_nullbyte: split the matched string on nullbyte and take only the first part
        """
        if not hasattr(self, "re_patterns"):
            logger.warning("No memory search performed (no pattern found)")
            return
        dump = self.memdump(verbose=False)
        with open(dump) as f:
            content = f.read()
        for pattern, fmt, descr in self.re_patterns:
            r = re.compile(pattern, re.M + re.S)
            out = r.search(content)
            if out is not None:
                out = out.group()
                #FIXME: split on nullbyte for each occurrence then join all
                if split_on_nullbyte:
                    out = out.split('\x00', 1)[0]
                self.save(out, self.result('memdump-{}'.format(descr), fmt))
        self.dump._artifacts.append(dump)
    
    def parse(self, output):
        """
        Dummy shortcut to self.dump.OutputParser.

        :param output: Volatility command output as a string
        """
        return self.dump.OutputParser(output)
    
    def result(self, cmd, fmt=None):
        """
        Compose the resource path depending on the executed command and given format.
        
        :param cmd: Volatility command name
        :param fmt: output format extension
        :return:    relative resource path
        """
        _ = self.name.lower() + "{}-{}{}"
        _ = _.format(["-{}".format(self.pid), ""][self.pid is None], cmd,
                     [".{}".format(fmt), ""][fmt is None])
        return join(self.dump.out_dir, _)

    def run(self):
        """
        Executes the 'memdump' Volatility command for extracting the related process memory.
        Public run method to be overridden for the operations to be applied to the related application.

        :param verbose: display the log message after saving the memory dump
        """
        raise NotImplementedError("Subclass GenericDumper and override run()")

    def save(self, content, dst, verbose=True, header=0):
        """
        Save a resource to a destination according to naming conventions.

        :param content: content to be saved
        :param dst:     relative resource path
        :param verbose: verbose mode
        :param header:  number of heading rows to be considered
        :return:        path where the file was saved
        """
        content = content or ""
        if not isinstance(content, str):
            raise ValueError("Bad content variable type (should be str)")
        if not isinstance(verbose, bool):
            raise ValueError("Bad verbose variable type (should be bool)")
        content = content.strip()
        if content == "" or len(content.split('\n')) == header:
            logger.debug("Empty content, no resource saved")
            return
        with open(dst, 'wb+') as f:
            f.write(content)
        if verbose:
            logger.info("> {}".format(relpath(dst)))
        return dst
    
    def shell(self, cmd, clean=False):
        """
        Executes an OS command.
        
        :param cmd: command line as text
        :return:    stdout, stderr
        """
        cmd = cmd.split()
        try:
            p = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
            out, err = p.communicate()
            if clean and isfile(cmd[-1]):
                os.remove(cmd[-1])
            return out, err
        except CalledProcessError:
            logger.error("Command '{}' failed".format(cmd))
            return None, None

    def vaddump(self, verbose=True):
        """
        Executes the 'vaddump' Volatility command for extracting the VAD tree for a given process.

        :param verbose: display the log message after grabbing the VAD nodes
        :return:        path to the VAD dump directory
        """
        if not isinstance(verbose, bool):
            raise ValueError("Bad verbose variable type (should be bool)")
        dst = self._makedir('vad')
        out = self.dump.call('vaddump', "-p {} -D {}".format(self.pid, dst))
        if verbose:
            logger.info("> {}".format(relpath(dst)))
        return dst

    def vadsearch(self, stop=True, include_pattern=False, reduce_text=False):
        """
        Executes the 'vaddump' Volatility command then parse the collected VAD nodes against used-defined patterns for
         collecting resources.
        Valid patterns structure:
          self.fmt_patterns := list(tuples(start_pattern, end_pattern, format))
        
        :param stop:            stop after the first resource matching the patterns
        :param include_pattern: include the found pattern when saving resource
        :param reduce_text:     use GenericDumper to reduce the output (makes sense if the output is text)
        """
        if not isinstance(stop, bool):
            raise ValueError("Bad stop variable type (should be bool)")
        if not isinstance(include_pattern, bool):
            raise ValueError("Bad include_pattern variable type (should be bool)")
        if not isinstance(reduce_text, bool):
            raise ValueError("Bad reduce_text variable type (should be bool)")
        if not hasattr(self, "fmt_patterns"):
            logger.warning("No VAD search performed (no pattern found)")
            return
        dst = self.vaddump(verbose=False)
        for fn in os.listdir(dst):
            with open(join(dst, fn), 'rb') as f:
                node = f.read()
            # choose the first start matching pattern in the provided list of format patterns
            found = False
            for start, end, fmt in self.fmt_patterns:
                try:
                    out = ["", start][include_pattern] + node.split(start, 1)[1]
                    out = out.split(end, 1)[0] + ["", end][include_pattern]
                    found = True
                    break
                except:
                    continue
            if found:
                if reduce_text:
                    out = GenericDumper.reduce_text(out)
                self.save(out.rstrip('\x00'), self.result('vadnode', fmt))
                if stop:
                    break
        shutil.rmtree(dst)

    def yarascan(self, pattern, verbose=True):
        """
        Executes the 'yarascan' Volatility command for matching a pattern from the related process memory.
        
        :param pattern: yara pattern
        :param verbose: display the log message after saving the scan result
        """
        if not isinstance(pattern, str):
            raise ValueError("Bad pattern variable type (should be str)")
        if not isinstance(verbose, bool):
            raise ValueError("Bad verbose variable type (should be bool)")
        cmd = 'yarascan'
        out = self.dump.call(cmd, "-p {} -Y '/{}/'".format(self.pid, pattern))
        if len(out.strip()) == 0:
            return
        i = 0
        dst = self.result(cmd + "-{:0<2}".format(i), "txt")
        while exists(dst):
            i += 1
            dst = self.result(cmd + "-{:0<2}".format(i), "txt")
        with open(dst, 'wb') as f:
            f.write(out)
        if verbose:
            logger.info("> {}".format(relpath(dst)))

    @staticmethod
    def reduce_text(text, alphabet=string.printable, wsize=5, threshold=3):
        """
        Determines start|end bounds based on a window of booleans telling if the characters are well in the allowed
         alphabet with the constraints that the first character in the window must be a not-allowed one and the number
         of not-allowed ones in the window is above the threshold.
        """
        if not isinstance(text, str):
            raise ValueError("Bad text variable type (should be str)")
        if not isinstance(alphabet, str):
            raise ValueError("Bad alphabet variable type (should be str)")
        if not isinstance(wsize, int):
            raise ValueError("Bad wsize variable type (should be int)")
        if not isinstance(threshold, int):
            raise ValueError("Bad threshold variable type (should be int)")
        # NB: the text is in UTF-16 little-endian and can simply be retrieved by removing the nullbytes after each
        #      normal character
        text = text.replace('\x00', '').replace('\r\n', '\n')
        halflen = len(text) / 2
        bounds = [halflen, len(text) - halflen - 1]
        for i, t in enumerate([text[:halflen][::-1], text[halflen:]]):
            w = deque(maxlen=wsize)
            bound_adapted = False
            # slide along the text to adapt the bound
            for k, c in enumerate(t):
                w.append(c not in alphabet)
                if w[0] and sum(w) >= threshold:
                    bounds[i] = k - len(w) + 1
                    bound_adapted = True
                    break
            # handle the window of the text tail if bound not adapted yet
            if not bound_adapted:
                for j in range(wsize - 1):
                    new_th = int(round(float(threshold) / wsize * len(w)))
                    if w.popleft() and sum(w) >= new_th:
                        bounds[i] = k - len(w)
                        break
        return text[halflen-bounds[0]:halflen+bounds[1]]




# -*- coding: UTF-8 -*-
"""AppmemDumper package information.

"""
import os

__all__ = ["__author__", "__copyright__", "__license__", "__version__"]
__author__    = "Alexandre D'Hondt"
__copyright__ = "A. D'Hondt"
__license__   = "agpl-3.0"

with open(os.path.join(os.path.dirname(__file__), "VERSION.txt")) as f:
    __version__ = f.read().strip()




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
from .application import *
from .system import *
from .template import *

from .application import __all__ as apps
from .system import __all__ as syst


APPDUMPERS = sorted(apps)
SYSDUMPERS = sorted(syst)
__all__ = ["APPDUMPERS", "SYSDUMPERS"] + APPDUMPERS + SYSDUMPERS




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
from .template import DumperTemplate


__all__ = [
    "AdobeReader",
    "Chrome",
    "Firefox",
    "FoxitReader",
    "InternetExplorer",
    "KeePass",
    "MediaPlayerClassic",
    "MSPaint",
    "Notepad",
    "OpenOffice",
    "PDFLite",
    "SumatraPDF",
    "Thunderbird",
    "TrueCrypt",
    "Wordpad",
]


class AdobeReader(DumperTemplate):
    """
    Dumper for the well-known application Adobe Reader. It uses the 'run()' method of DumperTemplate to extract the
     memory of the process and carves PDF files on it with Foremost then removes the memory dump.
    """
    procnames = ["reader.exe", "AcroRd32.exe"]
    only_parent = True

    def run(self):
        """
        Executes the 'memdump' Volatility command (DumperTemplate), carves PDF files with Foremost then removes the
         process memory dump.
        """
        self.carve("pdf", clean=True)


class Chrome(DumperTemplate):
    """
    Dumper for the common application Google Chrome.
    """
    procnames = ["chrome.exe"]

    def run(self):
        """
        Executes the Chrome-related Volatility community plugins.
        """
        self.commands('chromehistory', 'chromevisits', 'chromedownloadchains', 'chromesearchterms', 'chromedownloads',
                      'chromecookies', options="--output=csv", failmode="warn")


class Firefox(DumperTemplate):
    """
    Dumper for the common application Mozilla Firefox.
    """
    procnames = ["firefox.exe"]
    only_parent = True

    def run(self):
        """
        Executes the Firefox-related Volatility community plugins.
        """
        self.commands('firefoxcookies', 'firefoxdownloads', 'firefoxhistory',
                      header=1, options="--output=csv", failmode="warn")


class FoxitReader(AdobeReader):
    """
    Dumper for the common application Foxit Reader. It performs the same operations as AdobeReaderDumper.
    """
    procnames = ["FoxitReader.ex", "FoxitReader.exe"]


class InternetExplorer(DumperTemplate):
    """
    Dumper for the common application Microsoft Internet Explorer.
    """
    procnames = ["iexplore.exe"]
    re_patterns = [(r'function FindProxyForURL.{232}', 'txt', 'proxy-rules'), (r'Client UrlCache}', 'txt', 'urlcache')]

    def run(self):
        """
        Executes some IExplorer-related Volatility command.
        """
        self.commands('iehistory')
        self.memsearch(split_on_nullbyte=True)


class KeePass(DumperTemplate):
    """
    Dumper for the common application KeePass.
    """
    procnames = ["KeePass.exe", "KeePassX.exe", "PassKeep.exe"]
    messages = ["If the KeePass database was used with a master password ;\n"
                "1. Its hash can be recovered using 'keepass2john'\n"
                "2. It can then be cracked with 'john'"]
    # https://github.com/Stoom/KeePass/wiki/KDBX-v2-File-Format
    fmt_patterns = [("\x03\xd9\xa2\x9a\x65\xfb\x4b\xb5", "\x00" * 16, "kdb"),
                    ("\x03\xd9\xa2\x9a\x66\xfb\x4b\xb5", "\x00" * 16, "kdbx"),
                    ("\x03\xd9\xa2\x9a\x67\xfb\x4b\xb5", "\x00" * 16, "kdbx")]
    re_patterns = [(r'(<\?xml(\s[a-z0-9\=\-\"\'\._]+)+\?>\r?\n<KeePassFile>(.*?)<\/KeePassFile>)', "xml", "file"),
                   (r'(<\?xml(\s[a-zA-Z0-9\=\-\"\'\.\:\/_]+)+\?>\r?\n<ArrayOfString(\s[a-zA-Z0-9\=\-\"\'\.\:\/_]+)+>'
                    r'(.*?)<\/ArrayOfString>)', "xml", "path")]

    def run(self):
        """
        Executes the 'memdump' Volatility command (DumperTemplate), retrieves some XML content, then executes the
         'vaddump' Volatility command and finally gets the KeePass DB from the VAD nodes.
        """
        self.memsearch()
        self.vadsearch(include_pattern=True)


class MediaPlayerClassic(DumperTemplate):
    """
    Dumper for the common application Media Player Classic. It performs the same operations as MSPaintDumper.
    """
    procnames = ["mpc-hc.exe"]
    messages = DumperTemplate._predef_messages[0:1]

    def run(self):
        self.memdump()


class MSPaint(DumperTemplate):
    """
    Dumper for the well-known application Paint built in Microsoft Windows. It uses the 'run()' method of DumperTemplate
     to extract the memory of the process for further analysis using the 'memdump' Volatility command.
    """
    procnames = ["mspaint.exe"]
    messages = DumperTemplate._predef_messages[0:1]

    def run(self):
        self.memdump()


class Notepad(DumperTemplate):
    """
    Dumper for the well-known application Notepad built in Microsoft Windows. It tries the 'notepad' plugin, then tries
     the 'editbox' plugin or finally gets the text contained in the edition control box of the main window of Notepad
     from a VAD node based on a pattern.
    """
    procnames = ["notepad.exe"]
    fmt_patterns = [("\xf2\xf3\xf3\xff\xf1\xf2\xf2\xff\xf0\xf1\xf0\xff\xf0\xf1\xf1\xff", "\x00" * 16, "txt")]

    def run(self):
        """
        Executes the 'editbox' Volatility command.
        """
        # Try 1: use 'notepad' command (only for profiles <= WinXP...
        if not hasattr(self, "_no_notepad"):
            cmd = 'notepad'
            try:
                out = self.execute(cmd, text=True).split("Text:\n", 1)[1]
                self.save(out, self.result(cmd))
                return
            except IndexError:
                self.logger.debug("Nothing found with 'notepad'")
            except:
                self.logger.debug("'notepad' is not supported for this profile")
                self._no_notepad = True
        # Try 2: use 'editbox'
        cmd = 'editbox'
        out = self.execute(cmd, text=True)
        out = out.split("******************************\n")
        for result in out[1:]:
            meta, text = result.split("-------------------------\n")
            if self.has(self.parse(meta)['Process ID']):
                self.save(text, self.result(cmd))
                return
        self.logger.debug("Nothing found with 'editbox'")
        # Try 3: use 'vaddump' and search for patterns in VAD nodes
        self.vadsearch(reduce_text=True)


class OpenOffice(DumperTemplate):
    """
    Dumper for the common OpenOffice suite.
    """
    procnames = ["soffice.exe", "soffice.bin", "swriter.exe", "scalc.exe", "simpress.exe", "sdraw.exe", "sbase.exe",
                 "smath.exe", "sweb.exe"]
    # https://ubuntuforums.org/showthread.php?t=1378119
    re_patterns = [(r'(PK).{28}mimetypeapplication/vnd.oasis.opendocument.textPK(.*?)META-INF/manifest.xmlPK.{20}',
                    "odt", "text"),
                   (r'(PK).{28}mimetypeapplication/vnd.oasis.opendocument.spreadsheetPK(.*?)META-INF/manifest.xmlPK'
                    r'.{20}', "ods", "spreadsheet"),
                   (r'(PK).{28}mimetypeapplication/vnd.oasis.opendocument.presentationPK(.*?)META-INF/manifest.xmlPK'
                    r'.{20}', "odp", "presentation"),
                   (r'(PK).{28}mimetypeapplication/vnd.oasis.opendocument.graphicsPK(.*?)META-INF/manifest.xmlPK.{20}',
                    "odg", "graphics"),
                   (r'(PK).{28}mimetypeapplication/vnd.oasis.opendocument.chartPK(.*?)META-INF/manifest.xmlPK.{20}',
                    "odc", "chart"),
                   (r'(PK).{28}mimetypeapplication/vnd.oasis.opendocument.formulaPK(.*?)META-INF/manifest.xmlPK.{20}',
                    "odf", "formula"),
                   (r'(PK).{28}mimetypeapplication/vnd.oasis.opendocument.imagePK(.*?)META-INF/manifest.xmlPK.{20}',
                    "odi", "image"),
                   (r'(PK).{28}mimetypeapplication/vnd.oasis.opendocument.text-masterPK(.*?)META-INF/manifest.xmlPK'
                    r'.{20}', "odm", "textmaster"),
                   (r'(PK).{28}mimetypeapplication/vnd.sun.xml.writerPK(.*?)META-INF/manifest.xmlPK.{20}',
                    "sxw", "writer")]

    def run(self):
        """
        Executes the 'memdump' Volatility command (DumperTemplate) and retrieves OpenOffice documents.
        """
        self.memsearch()


class PDFLite(AdobeReader):
    """
    Dumper for the common application PDFLite. It performs the same operations as AdobeReaderDumper.
    """
    procnames = ["PDFlite.exe"]


class SumatraPDF(AdobeReader):
    """
    Dumper for the common application Sumatra PDF. It performs the same operations as AdobeReaderDumper.
    """
    procnames = ["SumatraPDF.exe"]


class Thunderbird(DumperTemplate):
    """
    Dumper for the common application Thunderbird.
    """
    procnames = ["thunderbird.ex", "thunderbird.exe"]
    messages = ["You can grep the .dmp file on common email-related keywords (e.g. From, To, ...)"]

    def run(self):
        """
        Executes the 'memdump' Volatility command.
        """
        self.memdump()


class TrueCrypt(DumperTemplate):
    """
    Dumper for the common application TrueCrypt.
    """
    procnames = ["TrueCrypt.exe"]

    def run(self):
        """
        Executes the 'memdump' Volatility command (DumperTemplate) and the TrueCrypt-related Volatility commands.
        """
        self.memdump()
        self.commands('truecryptmaster', 'truecryptpassphrase', 'truecryptsummary')


class Wordpad(DumperTemplate):
    """
    Dumper for the well-known application Wordpad built in Microsoft Windows. It extracts process' memory using the
     'memdump' command. It uses the 'carve()' method of DumperTemplate to extract image resources and executes the
     'vaddump' command for retrieving Virtual Address Descriptor (VAD) objects for later manual analysis.
    """
    procnames = ["wordpad.exe"]
    messages = DumperTemplate._predef_messages[0:2]

    def run(self):
        """
        Executes the 'memdump' Volatility command (DumperTemplate), carves files with Foremost then executes the
         'editbox' Volatility command.
        """
        self.carve("jpg", "png", clean=True)
        # TODO: find patterns for searching into the VAD nodes
        self.vaddump()




#!/usr/bin/python
# -*- coding: UTF-8 -*-
import logging
import os
import shlex
import shutil
import StringIO
import sys
from copy import deepcopy
from os.path import abspath, dirname, exists, expanduser, isdir, isfile, join
from subprocess import check_output, CalledProcessError, PIPE, Popen

import volatility.addrspace as addrspace
import volatility.conf as conf
import volatility.commands as commands
import volatility.constants as constants
import volatility.debug as debug
import volatility.obj as obj
import volatility.registry as registry
import volatility.utils as utils

from .dumpers import *

try:
    Popen(["foremost", "-h"], stdout=PIPE, stderr=PIPE)
except OSError:
    print("Missing dependencies, please run 'sudo apt-get install foremost'")
    sys.exit(1)


__all__ = ["APPDUMPERS", "SYSDUMPERS", "VolatilityMemDump"]
CMD = 'volatility {opt} {cmd}'
logger = logging.getLogger("main")


is_memdump = lambda x: isinstance(x, VolatilityMemDump)


class VolatilityAppDumper(object):
    """
    Volatility application dumper "relay" class for choosing the right application dumper based on the process name. It
     is called by VolatilityMemDump at process list creation to attach the desired dumpers.
    It instantiates a child class of GenericDumper adapted to the given application name.
    """
    def __init__(self, dump, pname=None):
        """
        Attaches the parent memory dump file and instantiates the right child class of GenericDumper adapted for the
         input application name.

        :param dump:  VolatilityMemDump instance
        :param pname: application process name
        """
        if not isinstance(dump, VolatilityMemDump):
            raise ValueError("Not a VolatilityMemDump instance")
        if pname is not None and not isinstance(pname, str):
            raise ValueError("Bad process name variable type (should be str)")
        # first, find if PID's exist when a process name is given
        pids = []
        if pname is not None:
            for p in dump.processes:
                if pname == str(p.ImageFileName).lower():
                    pids.append(str(p.UniqueProcessId))
            if len(pids) == 0:
                logger.error("No PID found for '{}'".format(pname))
                return
        # second, find the right application dumper for the given process name and attach the dumper to the memory dump
        #  if a dumper is found and if the related application is handled
        # if no process name was given and there exists a system dumper (that has no list of process names), create a
        #  dumper
        for cls in APPDUMPERS + SYSDUMPERS:
            pnames = globals()[cls].procnames
            if pnames is not None:
                pnames = [p.lower() for p in pnames]
            if (pname is None and pnames is None) or (pname is not None and pnames is not None and pname in pnames):
                if cls in dump._selected_apps or cls in dump._selected_syst:
                    if pname is not None:
                        dump._stats[0].append(pname)  # found
                    self.dumper = globals()[cls](dump, pids)
                    dump.dumpers[cls] = self.dumper
                else:
                    if pname is not None:
                        dump._stats[1].append(pname)  # not selected
                if pname is not None and pnames is not None:
                    return
        if pname is not None:
            dump._stats[2].append(pname)              # not handled


class VolatilityMemDump(object):
    """
    Volatility memory dump class that handles profile search and processes list retrieval. It allows either to execute a
     command from the Python API or to call Volatility in a subprocess.
    At processes list creation, it creates a list of Volatility application dumpers for the common applications handled
     by this tool by instantiating VolatilityAppDumper. It also provides a 'dump()' method to trigger the dumping of
     resources with the application dumpers.
    """
    class OutputParser(dict):
        """
        Dictionary that parses the output of a Volatility command at creation with:
         - every colon-separated parameter as a (key, value) pair
         - other lines under the 'others' key
        """
        def __init__(self, output):
            """
            Dictionary constructor that parses the Volatility command output.

            :param output: Volatility command output as a string
            """
            if not isinstance(output, str):
                raise ValueError("Bad output variable type (should be str)")
            self['others'] = []
            for line in output.split('\n'):
                line = line.strip().split(':', 1)
                if len(line) == 1:
                    if line[0].strip() != '':
                        self['others'].append(line[0].strip())
                else:
                    self[line[0].strip()] = line[1].strip()
    parsers = {
        'imageinfo': lambda o: VolatilityMemDump.OutputParser(o) ["Suggested Profile(s)"].split(", "),
    }

    def __init__(self, dump, apps, syst, out_dir="files", plugins_dir=None, profile=None, from_cache=True):
        """
        Determines the profile, retrieves the list of processes and creates the list of application dumpers.

        :param dump:        Volatility memory dump filename
        :param apps:        list of the application dumper classes to be handled
        :param syst:        list of the system dumper classes to be handled
        :param out_dir:     output directory for retrieved resources
        :param plugins_dir: Volatility custom plugins directory
        :param profile:     forced Volatility profile (preceeds from_cache)
        :param from_cache:  indicates if the profile must be retrieved from a cache file in /tmp or found by Volatility
        """
        short = dump
        dump = abspath(dump)
        if not isfile(dump):
            raise ValueError("{} is not a dump file".format(dump))
        if any(x not in APPDUMPERS for x in apps):
            raise ValueError("Unknown application dumper(s)")
        if any(x not in SYSDUMPERS for x in syst):
            raise ValueError("Unknown system dumper(s)")
        if plugins_dir is not None and not isdir(plugins_dir):
            raise ValueError("Bad plugins dir")
        if not isinstance(from_cache, bool):
            raise ValueError("Bad from-cache variable type (should be bool)")
        self._artifacts = []
        self._cache = {}
        self._selected_apps = apps
        self._selected_syst = syst
        if len(self._selected_apps) == 0 and len(self._selected_syst) == 0:
            logger.warning("No dumper selected")
            sys.exit(0)
        logger.debug("Setting output directory to '{}'...".format(out_dir))
        self.out_dir = abspath(out_dir)
        if not exists(self.out_dir):
            os.makedirs(self.out_dir)
        self._cachefile = join(self.out_dir, ".cache")
        # initialize dump opening
        registry.PluginImporter()
        self.__is_profile_tested = False
        self.config = conf.ConfObject()
        if plugins_dir is not None:
            plugins_dir = expanduser(plugins_dir)
            logger.debug("Setting plugins directory to '{}'...".format(plugins_dir))
            self.config.plugins = abspath(plugins_dir)
        for cls in [commands.Command, addrspace.BaseAddressSpace]:
            registry.register_global_options(self.config, cls)
        self.__commands = {k.lower(): v for k, v in registry.get_plugin_classes(commands.Command).items()}
        self.config.LOCATION = "file://{}".format(dump)
        # get the right dump profile and test it while getting processes
        logger.info("Opening dump file '{}'...".format(short))
        self.__is_profile_tested = self.__get_profile(profile, from_cache)

    def __get_profile(self, profile=None, from_cache=True):
        """
        Get the dump image profile by loading it from a cache file or by running 'imageinfo' Volatility command.

        :param profile:    force profile value (has the precedence on from_cache)
        :param from_cache: load the profile from the cache file if it exists
        :return: boolean   indicating if the profile is tested as correct
        """
        if not isinstance(from_cache, bool):
            raise ValueError("From-cache variable should be a boolean")
        cf = self._cachefile
        logger.info("Getting profile...")
        # get available profiles and commands
        available_profiles = self._get_available_profiles()
        if profile is not None and profile not in available_profiles.keys():
            raise ValueError("Bad Volatility profile '{}'".format(profile))
        compatible_profiles = []
        # load profile from 'profile' variable if not None or from cache, or use
        #  the 'imageinfo' command to find it
        self.config.PROFILE = None
        if profile is not None or from_cache:
            if profile is not None:
                self.config.PROFILE = profile
            elif from_cache and isfile(cf):
                with open(cf) as f:
                    self.config.PROFILE = f.readlines()[0].strip()
            if self.config.PROFILE is not None:
                if not self.__get_pslist():
                    self.config.PROFILE = None
                else:
                    compatible_profiles = [self.config.PROFILE]
        if self.config.PROFILE is None:
            self.config.PROFILE = "WinXPSP2x86"  # reset to default
            compatible_profiles = [p for p in self.call('imageinfo', parser=self.parsers['imageinfo'])]
        self.__profiles = {p: c for p, c in available_profiles.items() if p in compatible_profiles}
        # try 'pslist' command on the found profile ; if an error occurs, try the next profile in the list
        cursor, is_profile_tested = 0, False
        while not is_profile_tested and cursor < len(self.__profiles):
            self.config.PROFILE = self.__profiles.keys()[cursor]
            logger.debug("Candidate profile: {}".format(self.config.PROFILE))
            is_profile_tested = self.__get_pslist()
            cursor += 1
        if not is_profile_tested:
            logger.error("No suitable profile found ; please check that your memory dump is supported by Volatility")
            sys.exit(2)
        self._update_cache(self.config.PROFILE)
        logger.info("> Selected profile: {}".format(self.config.PROFILE))
        return is_profile_tested

    def __get_pslist(self):
        """
        Run 'pslist' Volatility command to check if the profile is correctly set, compute the list of processes and
         attach dumper objects for known applications.

        :return: boolean telling if the profile is correct
        """
        if hasattr(self, "processes"):
            return True
        logger.info("Getting processes...")
        try:
            self.processes = self.execute('pslist')
            self.processes.extend(self.execute("psscan"))
        except:
            logger.warning("Wrong profile, re-getting profile...")
            return False
        # collect PIDs relationships
        procs = {'0': "Init"}
        for p in self.processes:
            procs[str(int(p.UniqueProcessId))] = p.ImageFileName
        self.ppids = {}
        for p in self.processes:
            pid = str(int(p.UniqueProcessId))
            ppid = str(int(p.InheritedFromUniqueProcessId))
            try:
                k = (ppid, procs[ppid])
            except KeyError:
                k = (ppid, None)
            self.ppids.setdefault(k, [])
            self.ppids[k].append((pid, procs[pid]))
        # parse the list of processes and create objects for each set of application instances that are handled by the
        #  present tool
        self.dumpers = {}
        self._stats = [[], [], []]
        procnames = set([str(p.ImageFileName) for p in self.processes])
        # first, search for system dumpers
        VolatilityAppDumper(self)
        # second, search for application dumpers per process name
        for n in sorted(procnames, key=lambda n: n.lower()):
            VolatilityAppDumper(self, n.lower())
        for i, t in enumerate(["Found", "Not selected", "Not handled"]):
            s = self._stats[i]
            if len(s) > 0:
                logger.debug("{}: {}".format(t.ljust(12), ", ".join(s)))
        if len(self.dumpers) == 0:
            logger.warning("No application to be handled")
        return True
    
    def _get_available_profiles(self):
        """
        Get the list of available profiles.
        """
        return registry.get_plugin_classes(obj.Profile)
    
    def _update_cache(self, profile=None, dumper=None):
        """
        Update cache file for the current memory dump.
        
        :param profile: profile to be written to the cache
        :param dumper:  new dumper to be added to the cache
        """
        if profile is None and dumper is None:
            return
        cf = self._cachefile
        if not exists(cf):
            with open(cf, 'w') as f:
                f.write(self.config.PROFILE)
        with open(cf) as f:
            c = f.read().splitlines()
        p, d = c[0], c[1:]
        if profile is not None:
            c[0] = profile
        if dumper is not None:
            d = sorted(list(set(d + [dumper])))
            c[1:] = d
        with open(cf, 'w') as f:
            f.write('\n'.join(c))

    def call(self, command, options=None, parser=None, failmode="silent"):
        """
        Run Volatility with given command and options using subprocess.

        :param command:  Volatility command name
        :param options:  valid command-related options
        :param parser:   None or function for parsing the output
        :param failmode: string indicating the fail mode (error|warn|silent)
        :return:         parsed output
        """
        if not isinstance(command, str):
            raise ValueError("Bad command variable type (should be str)")
        if options is not None and not isinstance(options, str):
            raise ValueError("Bad options variable type (should be str)")
        if parser is not None and not callable(parser):
            raise ValueError("Bad parser variable type (should be callable)")
        if failmode not in ["error", "warn", "silent"]:
            raise ValueError("Bad failmode value (should be error|warn|silent)")
        # compose the list of options with already known and input information
        options = [options] if options is not None else []
        if self.__is_profile_tested:
            options.insert(0, "--profile={}".format(self.config.PROFILE))
        filepath = self.config.LOCATION.split("://", 1)[1]
        options.insert(0, "--file=\"{}\"".format(filepath))
        if self.config.plugins is not None and len(self.config.plugins) > 0:
            options.insert(0, "--plugins={}".format(self.config.plugins))
        cmd = CMD.format(cmd=command, opt=" ".join(options))
        # try to get the result from cache first
        if cmd in self._cache:
            logger.debug("> Got result of command '{}' from cache...".format(command))
            logger.debug(">> {}".format(cmd))
            return self._cache[cmd]
        # run the resulting command with a subprocess
        logger.debug("> Calling command '{}'...".format(command))
        logger.debug(">> {}".format(cmd))
        p = Popen(shlex.split(cmd), stdin=PIPE, stdout=PIPE, stderr=PIPE)
        out, err = p.communicate()
        if "ERROR" in err:
            if failmode == "silent":
                return
            if "You must specify something to do (try -h)" in err:
                msg = "Plugin '{}' not found".format(command)
            else:
                msg = "({}) {}".format(*map(lambda x: x.strip(), err.split(":", 2)[1:3]))
            if failmode == "warn":
                logger.warning(msg)
                return
            elif failmode == "error":
                logger.error(msg)
                raise CalledProcessError(1, cmd)
        out = out.replace('\r\n', '\n')
        # then parse the output with the given parser and return the result
        if parser is not None:
            out = parser(out)
        # cache the result
        self._cache[cmd] = out
        return out
    
    def clean(self):
        """
        Clean up artifacts registered in self._artifacts.
        """
        for a in self._artifacts:
            try:
                os.remove(a)
                logger.debug("Removed '{}'".format(a))
            except OSError:
                logger.debug("Could not remove '{}'".format(a))

    def dump(self, update=False):
        """
        Dump resources for every known application in the memory dump.
        
        :param update: update output directory (meaning that collected pieces of information should not be recollected
                        using Volatility commands)
        """
        if not update and exists(self.out_dir):
            shutil.rmtree(self.out_dir) if isdir(self.out_dir) else os.remove(self.out_dir)
        elif update and isfile(self.out_dir):
            logger.error("Output directory is not valid")
            sys.exit(2)
        if not exists(self.out_dir):
            os.makedirs(self.out_dir)
        # run each dumper and collect information messages for guidance
        d = self.dumpers
        logger.info("Processing {} dumper{}{}...".format(len(d), ["", "s"][len(d) > 1],
                                                         ["", ", this may take a while"][len(d) > 3]))
        msgs = {}
        for n, a in d.items():
            logger.debug("Processing dumper '{}'...".format(n))
            for msg in a._run(update):
                msgs.setdefault(msg, [])
                msgs[msg].append(n)
        # finally, warn the collected messages
        for m, n in msgs.items():
            logger.warning("\nThe following applies to collected objects of:\n- {}\n\n{}\n".format("\n- ".join(n), m))
    
    def execute(self, command, nopid=False, text=False, parser=None):
        """
        Run Volatility command through the dedicated Python API.

        :param command: Volatility command name
        :param nopid:   ensure no PID in the config object (for some commands that search for application artifacts
                         without requiring a PID)
        :param text:    convert the result into text
        :param parser:  None or function for parsing the output
        :return:        parsed output
        """
        if command not in self.__commands.keys():
            raise ValueError("Command does not exist")
        if not isinstance(nopid, bool):
            raise ValueError("Bad nopid variable type (should be bool)")
        if not isinstance(text, bool):
            raise ValueError("Bad text variable type (should be bool)")
        if parser is not None and not callable(parser):
            raise ValueError("Bad parser variable type (should be callable)")
        # try to get the result from cache first
        if command in self._cache:
            logger.debug("> Got result of command '{}' from cache...".format(command))
            return self._cache[command]
        # run the command with the API
        logger.debug("> Executing command '{}'...".format(command))
        if nopid:
            pid = self.config.PID
            self.config.PID = None
        command = self.__commands[command](deepcopy(self.config))
        result = command.calculate()
        # convert the result to text if required
        if text:
            io = StringIO.StringIO()
            command.render_text(io, result)
            out = io.getvalue().replace('\r\n', '\n')
        else:
            out = list(result)
        if nopid:
            self.config.PID = pid
        # then parse the output with the given parser and return the result
        if parser:
            out = parser(out)
        # cache the result
        self._cache[command] = out
        return out

